import numpy as np
import math

from .utils import (
    MinCurve,
    get_nev,
    get_vec,
    get_angles,
    HLA_to_NEV,
    NEV_to_HLA,
    get_sigmas
)

from welleng.error import ErrorModel
from welleng.exchange.wbp import TurnPoint

class Survey:
    def __init__(
        self,
        md,
        inc,
        azi,
        n=None,
        e=None,
        tvd=None,
        x=None,
        y=None,
        z=None,
        vec=None,
        radius=None,
        cov_nev=None,
        cov_hla=None,
        error_model=None,
        well_ref_params=None,
        start_xyz=[0,0,0],
        start_nev=[0,0,0],
        start_cov_nev=None,
        deg=True,
        unit="meters"
    ):
        """
        Initialize a welleng.Survey object.

        Parameters
        ----------
            md: (,n) list or array of floats
                List or array of well bore measured depths.
            inc: (,n) list or array of floats
                List or array of well bore survey inclinations
            azi: (,n) list or array of floats
                List or array of well bore survey azimuths
            n: (,n) list or array of floats (default: None)
                List or array of well bore northings
            e: (,n) list or array of floats (default: None)
                List or array of well bore eastings
            tvd: (,n) list or array of floats (default: None)
                List or array of local well bore z coordinates, i.e. depth
                and usually relative to surface or mean sea level.
            x: (,n) list or array of floats (default: None)
                List or array of local well bore x coordinates, which is
                usually aligned to the east direction.
            y: (,n) list or array of floats (default: None)
                List or array of local well bore y coordinates, which is
                usually aligned to the north direction.
            z: (,n) list or array of floats (default: None)
                List or array of well bore true vertical depths relative
                to the well surface datum (usually the drill floor 
                elevation DFE, so not always identical to tvd).
            vec: (n,3) list or array of (,3) floats (default: None)
                List or array of well bore unit vectors that describe the
                inclination and azimuth of the well relative to (x,y,z)
                coordinates.
            radius: float or (,n) list or array of floats (default: None)
                If a single float is specified, this value will be
                assigned to the entire well bore. If a list or array of
                floats is provided, these are the radii of the well bore.
                If None, a well bore radius of 12" or approximately 0.3 m
                is applied.
            cov_nev: (n,3,3) list or array of floats (default: None)
                List or array of covariance matrices in the (n,e,v)
                coordinate system.
            cov_hla: (n,3,3) list or array of floats (default: None)
                List or array of covariance matrices in the (h,l,a)
                well bore coordinate system (high side, lateral, along
                hole).
            error_model: str (default: None)
                If specified, this model is used to calculate the
                covariance matrices if they are not present. Currently,
                only the "ISCWSA_MWD" model is provided.
            well_ref_params: dict (default: None)
                If an error_model is set, these well reference params
                are provided to the welleng.error.ErrorModel class. The
                defaults are:
                    dict(
                        Latitude = -40,     # degrees
                        G = 9.80665,        # m/s2
                        BTotal = 61000,     # nT
                        Dip = -70,          # degrees
                        Declination = 13,   # degrees  
                        Convergence = 0,    # degrees
                    )
            start_xyz: (,3) list or array of floats (default: [0,0,0])
                The start position of the well bore in (x,y,z) coordinates.
            start_nev: (,3) list or array of floats (default: [0,0,0])
                The start position of the well bore in (n,e,v) coordinates.
            start_cov_nev: (,3,3) list or array of floats (default: None)
                The covariance matrix for the start position of the well
                bore in (n,e,v) coordinates.
            deg: boolean (default: True)
                Indicates whether the provided angles are in degrees
                (True), else radians (False).
            unit: str (default: 'meters')
                Indicates whether the provided lengths and distances are
                in 'meters' or 'feet', which impacts the calculation of
                the dls (dog leg severity).

        Returns
        -------
        A welleng.survey.Survey object.
        """
        self.unit = unit
        self.deg = deg
        self.start_xyz = start_xyz
        self.start_nev = start_nev
        self.md = np.array(md)
        self.start_cov_nev = start_cov_nev
        self._make_angles(inc, azi, deg)
        self._get_radius(radius)
        
        self.survey_deg = np.array([self.md, self.inc_deg, self.azi_deg]).T
        self.survey_rad = np.array([self.md, self.inc_rad, self.azi_rad]).T

        self.n = n
        self.e = e
        self.tvd = tvd
        self.x = x
        self.y = y
        self.z = z
        self.vec = vec

        self._min_curve()
        self._get_toolface_and_rates()

        # initialize errors
        # TODO: read this from a yaml file in errors
        error_models = ["ISCWSA_MWD"]
        if error_model is not None:
            assert error_model in error_models, "Unrecognized error model"
        self.error_model = error_model
        self.well_ref_params = well_ref_params

        self.cov_hla = cov_hla
        self.cov_nev = cov_nev

        self._get_errors()

    def _get_radius(self, radius=None):
        if radius is None:
            self.radius = np.full_like(self.md.astype(float), 0.3048)
        elif np.array([radius]).shape[-1] == 1:
            self.radius = np.full_like(self.md.astype(float), radius)
        else:
            assert len(radius) == len(self.md), "Check radius"
            self.radius = np.array(radius)
    
    def _min_curve(self):
        """
        Get the (x,y,z), (n,e,v), doglegs, rfs, delta_mds, dlss and
        vectors for the well bore if they were not provided, using the
        minimum curvature method.
        """
        mc = MinCurve(self.md, self.inc_rad, self.azi_rad, self.start_xyz, self.unit)
        self.dogleg = mc.dogleg
        self.rf = mc.rf
        self.delta_md = mc.delta_md
        self.dls = mc.dls
        self.pos = mc.poss

        if self.x is None:
            # self.x, self.y, self.z = (mc.poss + self.start_xyz).T
            self.x, self.y, self.z = (mc.poss).T
        if self.n is None:
            self._get_nev()
        if self.vec is None:
            self.vec = get_vec(self.inc_rad, self.azi_rad, deg=False)

    def _get_nev(self):
        self.n, self.e, self.tvd = get_nev(
            np.array([
                self.x,
                self.y,
                self.z
            ]).T,
            start_xyz=self.start_xyz,
            start_nev=self.start_nev
        ).T

    def _make_angles(self, inc, azi, deg=True):
        """
        Calculate angles in radians if they were provided in degrees or
        vice versa.
        """
        if deg:
            self.inc_rad = np.radians(inc)
            self.azi_rad = np.radians(azi)
            self.inc_deg = np.array(inc)
            self.azi_deg = np.array(azi)
        else:
            self.inc_rad = np.array(inc)
            self.azi_rad = np.array(azi)
            self.inc_deg = np.degrees(inc)
            self.azi_deg = np.degrees(azi)

    def _get_errors(self):
        """
        Initiate a welleng.error.ErrorModel object and calculate the
        covariance matrices with the specified error model.
        """
        if self.error_model:
            if self.error_model == "ISCWSA_MWD":
                if self.well_ref_params is None:
                    self.err = ErrorModel(
                        survey=self.survey_deg,
                        surface_loc=self.start_xyz,
                    )
                else:
                    self.err = ErrorModel(
                        survey=self.survey_deg,
                        surface_loc=self.start_xyz,
                        well_ref_params=self.well_ref_params,
                    )
                self.cov_hla = self.err.errors.cov_HLAs.T
                self.cov_nev = self.err.errors.cov_NEVs.T
        else:
            if self.cov_nev is not None and self.cov_hla is None:
                self.cov_hla = NEV_to_HLA(self.survey_rad, self.cov_nev.T).T
            elif self.cov_nev is None and self.cov_hla is not None:
                self.cov_nev = HLA_to_NEV(self.survey_rad, self.cov_hla.T).T
            else:
                pass
        
        if (
            self.start_cov_nev is not None
            and self.cov_nev is not None
        ):
            self.cov_nev += self.start_cov_nev
            self.cov_hla = NEV_to_HLA(self.survey_rad, self.cov_nev.T).T

    def _curvature_to_rate(self, curvature):
        with np.errstate(divide='ignore', invalid='ignore'):
            radius = 1 / curvature
        circumference = 2 * np.pi * radius
        if self.unit == 'meters':
            x = 30
        else:
            x = 100
        rate = np.absolute(np.degrees(2 * np.pi / circumference) * x)

        return rate
    
    def _get_toolface_and_rates(self):
        """
        Reference SPE-84246.
        theta is inc, phi is azi
        """
        # split the survey
        s = SplitSurvey(self)

        if self.unit == 'meters':
            x = 30
        else:
            x = 100

        # this is lazy I know, but I'm using this mostly for flags
        with np.errstate(divide='ignore', invalid='ignore'):
            t1 = np.arctan(
                np.sin(s.inc2) * np.sin(s.delta_azi) /
                (
                    np.sin(s.inc2) * np.cos(s.inc1) * np.cos(s.delta_azi)
                    - np.sin(s.inc1) * np.cos(s.inc2)
                )
            )
            t1 = np.nan_to_num(
                np.where(t1 < 0, t1 + 2 * np.pi, t1),
                nan=np.nan
            )
            t2 = np.arctan(
                np.sin(s.inc1) * np.sin(s.delta_azi) /
                (
                    np.sin(s.inc2) * np.cos(s.inc1)
                    - np.sin(s.inc1) * np.cos(s.inc2) * np.cos(s.delta_azi)
                )
            )
            t2 = np.nan_to_num(
                np.where(t2 < 0, t2 + 2 * np.pi, t2),
                nan=np.nan
            )
            self.curve_radius = (360 / self.dls * x) / (2 * np.pi)
        
            curvature_dls = 1 / self.curve_radius

            self.toolface = np.concatenate((t1, np.array([t2[-1]])))

            curvature_turn = curvature_dls * (np.sin(self.toolface) / np.sin(self.inc_rad))
            self.turn_rate = self._curvature_to_rate(curvature_turn)

            curvature_build = curvature_dls * np.cos(self.toolface)
            self.build_rate = self._curvature_to_rate(curvature_build)

        # calculate plan normals
        n12 = np.cross(s.vec1, s.vec2)
        with np.errstate(divide='ignore', invalid='ignore'):
            self.normals = n12 / np.linalg.norm(n12, axis=1).reshape(-1,1)

def interpolate_survey(survey, x=0, index=0):
    """
    Interpolates a point distance x between two survey stations
    using minimum curvature.

    Parameters
    ----------
        survey: welleng.Survey
            A survey object with at least two survey stations.
        x: float
            Length along well path from indexed survey station to
            perform the interpolate at. Must be less than length
            to the next survey station.
        index: int
            The index of the survey station from which to interpolate
            from.

    Returns
    -------
        survey: welleng.Survey
            A survey object consisting of the two survey stations
            between which the interpolation has been made (index 0 and
            2), with the interpolated station between them (index 1)

    """
    index = int(index)

    assert index < len(survey.md) - 1, "Index is out of range"

    # assert x <= survey.delta_md[index + 1], "x is out of range"


    # check if it's just a tangent section
    if survey.dogleg[index + 1] == 0:
        azi = survey.azi_rad[index]
        inc = survey.inc_rad[index]

    else:
        # get the vector
        t1 = survey.vec[index]
        t2 = survey.vec[index + 1]

        total_dogleg = survey.dogleg[index + 1]

        dogleg = x * (survey.dogleg[index + 1] / survey.delta_md[index + 1])

        t = (
            (math.sin(total_dogleg - dogleg) / math.sin(total_dogleg)) * t1
            + (math.sin(dogleg) / math.sin(total_dogleg)) * t2
        )

        inc, azi = get_angles(t)[0]

    s = Survey(
        md=np.array([survey.md[index], survey.md[index] + x]),
        inc=np.array([survey.inc_rad[index], inc]),
        azi=np.array([survey.azi_rad[index], azi]),
        start_xyz=np.array([survey.x, survey.y, survey.z]).T[index],
        start_nev=np.array([survey.n, survey.e, survey.tvd]).T[index],
        deg=False
    )
    s._min_curve()

    return s

def slice_survey(survey, start, stop=None):
    """
    Take a slice from a welleng.survey.Survey object.

    Parameters
    ----------
        survey: welleng.survey.Survey object
        start: int
            The start index of the desired slice.
        stop: int (default: None)
            The stop index of the desired slice, else the remainder of
            the well bore TD is the default.

    Returns
    -------
        s: welleng.survey.Survey object
            A survey object of the desired slice is returned.
    """
    if stop is None:
        stop = start + 2
    md, inc, azi = survey.survey_rad[start:stop].T
    nevs = np.array([survey.n, survey.e, survey.tvd]).T[start:stop]
    n, e, tvd = nevs.T
    vec = survey.vec[start:stop]
        
    s = Survey(
        md=md,
        inc=inc,
        azi=azi,
        n=n,
        e=e,
        tvd=tvd,
        radius=survey.radius[start:stop],
        cov_hla=survey.cov_hla[start:stop],
        cov_nev=survey.cov_nev[start:stop],
        start_nev=[n[0], e[0], tvd[0]],
        deg=False,
        unit=survey.unit,
    )

    return s

def make_cov(a, b, c, diag=False):
    """
    Make a covariance matrix from the 1-sigma errors.

    Parameters
    ----------
        a: (,n) list or array of floats
            Errors in H or N/y axis.
        b: (,n) list or array of floats
            Errors in L or E/x axis.
        c: (,n) list or array of floats
            Errors in A or V/TVD axis.
        diag: boolean (default=False)
            If true, only the lead diagnoal is calculated
            with zeros filling the remainder of the matrix.

    Returns
    -------
        cov: (n,3,3) np.array
    """

    if diag:
        z = np.zeros_like(np.array([a]).reshape(-1))
        cov = np.array([
            [a * a, z, z],
            [z, b * b, z],
            [z, z, c * c]
        ]).T
    else:
        cov = np.array([
            [a * a, a * b, a * c],
            [a * b, b * b, b * c],
            [a * c, b * c, c * c]
        ]).T

    return cov

def make_long_cov(arr):
    """
    Make a covariance matrix from the half covariance 1sigma data.
    """
    aa, ab, ac, bb, bc, cc = np.array(arr).T
    cov = np.array([
        [aa, ab, ac],
        [ab, bb, bc],
        [ac, bc, cc]
    ]).T

    return cov

class SplitSurvey:
    def __init__(
        self,
        survey,
    ):
        self.md1, self.inc1, self.azi1 = survey.survey_rad[:-1].T
        self.md2, self.inc2, self.azi2 = survey.survey_rad[1:].T
        self.delta_azi = self.azi2 - self.azi1
        self.delta_inc = self.inc2 - self.inc1

        self.vec1 = survey.vec[:-1]
        self.vec2 = survey.vec[1:]
        self.dogleg = survey.dogleg[1:]   

def get_circle_radius(survey, **targets):
    # TODO: add target data to sections
    ss = SplitSurvey(survey)

    x1, y1, z1 = np.cross(ss.vec1, survey.normals).T
    x2, y2, z2 = np.cross(ss.vec2, survey.normals).T

    b1 = np.array([y1, x1, z1]).T
    b2 = np.array([y2, x2, z2]).T
    nev = np.array([survey.n, survey.e, survey.tvd]).T

    cc1 = (
        nev[:-1] - b1
        / np.linalg.norm(b1, axis=1).reshape(-1,1)
        * survey.curve_radius[:-1].reshape(-1,1)
    )
    cc2 = (
        nev[1:] - b2
        / np.linalg.norm(b2, axis=1).reshape(-1,1)
        * survey.curve_radius[1:].reshape(-1,1)
    )

    starts = np.vstack((cc1, cc2))
    ends = np.vstack((nev[:-1], nev[1:]))

    n = 1


    return (starts, ends)

def get_sections(survey, rtol=1e-1, atol=1e-2, **targets):
    """
    Tries to discretize a survey file into hold or curve sections. These
    sections can then be used to generate a WellPlan object to generate a
    .wbp format file for import into Landmark COMPASS, thus converting a
    survey file to an editable well trajectory.

    Note that this is in development and only tested on output from planning
    software. In its current form it likely won't be too successful on
    "as drilled" surveys (but optimizing the tolerances may help).

    Parameters
    ----------
        survey: welleng.survey.Survey object
        rtol: float (default: 1e-1)
            The relative tolerance when comparing the normals using the
            numpy.isclose() function.
        atol: float (default: 1e-2)
            The absolute tolerance when comparing the normals using the
            numpy.isclose() function.
        **targets: list of Target objects
            Not supported yet...

    Returns:
        sections: list of welleng.exchange.wbp.TurnPoint objects
    """

    # TODO: add target data to sections
    ss = SplitSurvey(survey)
    
    continuous = np.all(
        np.isclose(
            survey.normals[:-1],
            survey.normals[1:],
            rtol=rtol, atol=atol,
            equal_nan=True
        ), axis=-1
    )

    ends = np.concatenate(
        (np.where(continuous == False)[0] + 1, np.array([len(continuous)-1]))
    )
    starts = np.concatenate(([0], ends[:-1]))
    
    actions = [
        "hold" if a else "curve"
        for a in np.isnan(survey.toolface[starts])
    ]

    sections = []
    tie_on = True
    for s, e, a in zip(starts, ends, actions):
        md = survey.md[s]
        inc = survey.inc_deg[s]
        azi = survey.azi_deg[s]
        x = survey.e[s]
        y = survey.n[s]
        z = -survey.tvd[s]
        location = [x, y, z]

        target = ""
        if survey.unit == 'meters':
            denominator = 30
        else:
            denominator = 100

        if a == "hold":
            dls = 0.0
            toolface = 0.0
            build_rate = 0.0
            turn_rate = 0.0
            method = ""
        else:
            method = "0"
            dls = survey.dls[e]
            toolface = np.degrees(survey.toolface[s])
            # looks like the toolface is in range -180 to 180 in the .wbp file
            toolface = toolface - 360 if toolface > 180 else toolface
            delta_md = survey.md[e] - md

            # TODO: should sum this line by line to avoid issues with long sections
            build_rate = abs(
                (survey.inc_deg[e] - survey.inc_deg[s])
                / delta_md * denominator
            )

            # TODO: should sum this line by line to avoid issues with long sections
            # need to be careful with azimuth straddling north
            delta_azi_1 = abs(survey.azi_deg[e] - survey.azi_deg[s])
            delta_azi_2 = abs(360 - delta_azi_1)
            delta_azi = min(delta_azi_1, delta_azi_2)
            turn_rate = delta_azi / delta_md * denominator

        section = TurnPoint(
            md=md,
            inc=inc,
            azi=azi,
            build_rate=build_rate,
            turn_rate=turn_rate,
            dls=dls,
            toolface=toolface,
            method=method,
            target=None,
            tie_on=tie_on,
            location=location
        )

        sections.append(section)

        # Repeat the first section so that creating .wbp works
        if tie_on:
            sections.append(section)
            sections[-1].tie_on = False

        tie_on = False
    
    return sections


import numpy as np
from numpy import sin, cos, tan, pi, sqrt
from .utils import MinCurve
from .errors.iscwsa_mwd import iscwsaMwd

# TODO: there's likely an issue with TVD versus TVDSS that
# needs to be resolved. This model assumes TVD relative to
# rig floor, but often a TVDSS is provided instead (with a
# negative value for rig floor elevation).

class ErrorModel():
    """
    A class to initiate the field parameters and error magnitudes
    for subsequent error calculations.
    """

    class Error:
        '''
        Standard components of a well bore survey error.
        '''
        def __init__(
            self,
            code,
            propagation,
            e_DIA,
            cov_DIA,
            e_NEV,
            e_NEV_star,
            sigma_e_NEV,
            cov_NEV
        ):
  
            self.code = code
            self.propagation = propagation
            self.e_DIA = e_DIA
            self.cov_DIA = cov_DIA
            self.e_NEV = e_NEV
            self.e_NEV_star = e_NEV_star
            self.sigma_e_NEV = sigma_e_NEV
            self.cov_NEV = cov_NEV

    def __init__(
        self,
        surface_loc = np.array([0.0, 0.0, 0.0]),    # [x, y, z] not [n,e,v]
        survey = np.array([0.0, 0.0, 0.0]),         # [md, inc, azi] in degrees
        # TODO: move these params to a yaml file
        well_ref_params=dict(
            Latitude = -40,                         # degrees
            G = 9.80665,                            # m/s2
            BTotal = 61000,                         # nT
            Dip = -70,                              # degrees
            Declination = 13,                       # degrees  
            Convergence = 0,                        # degrees
        ),
        AzimuthRef = True,
        DepthUnits = 'meters',
        VerticalIncLimit = 0.0001,                  # degrees
        FeetToMeters = 0.3048,
        # TODO: move these params to a yaml file in the errors foler since
        # they're specific to an error model.
        errors_mag = dict(
            DRFR = 0.35,
            DSFS = 0.00056,
            DSTG = 2.5e-7,
            AB = 0.004,
            AS = 0.0005,
            MB = 70,
            MS = 0.0016,
            DECG = np.radians(0.36),
            DECR = np.radians(0.1),
            DBHG = np.radians(5000),
            DBHR = np.radians(3000),
            AMIL = 220,
            SAG = np.radians(0.2),
            XYM = np.radians(0.1)
        ),
        error_model="ISCWSA_MWD",
    ):
        """

        """

        error_models = [
            "ISCWSA_MWD"
        ]
        assert error_model in error_models, "Unrecognized error model"
        self.error_model = error_model

        p = well_ref_params

        self.surface_loc = surface_loc
        self.Latitude = np.radians(p["Latitude"])
        self.G = p["G"]
        self.BTotal = p["BTotal"]
        self.Dip = np.radians(p["Dip"])
        self.Declination = np.radians(p["Declination"])
        self.Convergence = np.radians(p["Convergence"])
        self.AzimuthRef = AzimuthRef
        self.DepthUnits = DepthUnits
        self.VerticalIncLimit = np.radians(VerticalIncLimit)
        self.FeetToMeters = FeetToMeters
        self.errors_mag = errors_mag

        self.survey_deg = np.array(survey)
        self.IncRad, self.AziRad = np.radians(self.survey_deg[:,1:]).T
        self.AziTrue = self.AziRad + np.full(len(self.AziRad), self.Convergence)
        self.AziG = self.AziTrue - np.full(len(self.AziRad), self.Convergence)
        self.AziMag = self.AziTrue - np.full(len(self.AziRad), self.Declination)

        self.survey_rad = np.stack((
            self.survey_deg[:,0],
            self.IncRad,
            self.AziTrue
        ), axis=-1)
        self.survey_drdp = self.survey_rad

        self.POS = [np.array(self.surface_loc)]
        self.DLS = [np.array(0)]
        md, inc, azi = self.survey_drdp.T
        self.mc = MinCurve(
            md=md,
            inc=inc,
            azi=azi,
            start_xyz=self.surface_loc,
            unit=self.DepthUnits
            )

        self.TVD = self.mc.poss[:,2]

        self.drdp = self._drdp(self.survey_drdp)
        self.drdp_sing = self._drdp_sing(self.survey_drdp)

        if self.error_model == "ISCWSA_MWD":
            self.errors = iscwsaMwd(self)

    def _e_NEV(self, e_DIA):
        D, I, A = e_DIA.T
        arr = np.array([
            (self.drdp[:,0] + self.drdp[:,9]) * D
            + (self.drdp[:,3] + self.drdp[:,12]) * I
            + (self.drdp[:,6] + self.drdp[:,15]) * A,

            (self.drdp[:,1] + self.drdp[:,10]) * D
            + (self.drdp[:,4] + self.drdp[:,13]) * I
            + (self.drdp[:,7] + self.drdp[:,16]) * A,

            (self.drdp[:,2] + self.drdp[:,11]) * D
            + (self.drdp[:,5] + self.drdp[:,14]) * I
            + (self.drdp[:,8] + self.drdp[:,17]) * A,
        ]).T

        arr[0] = 0

        return arr

    def _e_NEV_star(self, e_DIA):
        D, I, A = e_DIA.T
        arr = np.array([
            self.drdp[:,0] * D
            + self.drdp[:,3] * I
            + self.drdp[:,6] * A,

            self.drdp[:,1] * D
            + self.drdp[:,4] * I
            + self.drdp[:,7] * A,

            self.drdp[:,2] * D
            + self.drdp[:,5] * I
            + self.drdp[:,8] * A
        ]).T

        arr[0] = 0

        return arr

    def _cov(self, arr):
        '''
        Returns a covariance matrix from an (n,3) array.
        '''
        x, y, z = np.array(arr).T
        return np.array([
            [x*x, x*y, x*z],
            [y*x, y*y, y*z],
            [z*x, z*y, z*z]
        ])

    def _sigma_e_NEV_systematic(self, e_NEV, e_NEV_star):
        return e_NEV_star + np.vstack((e_NEV[0], np.cumsum(e_NEV, axis=0)[:-1]))

    def _generate_error(self, code, e_DIA, propagation='systematic', NEV=True, e_NEV=None, e_NEV_star=None):
        if not NEV:
            return e_DIA
        else:   
            cov_DIA = self._cov(e_DIA)
            if e_NEV is None:
                e_NEV = self._e_NEV(e_DIA)
                e_NEV_star = self._e_NEV_star(e_DIA)
            if propagation == 'systematic':
                sigma_e_NEV = self._sigma_e_NEV_systematic(e_NEV, e_NEV_star)
                cov_NEV = self._cov(sigma_e_NEV)
            elif propagation == 'random':
                sigma_e_NEV = np.cumsum(self._cov(e_NEV), axis=-1)
                cov_NEV = np.add(
                    self._cov(e_NEV_star),
                    np.concatenate((np.array(np.zeros((3,3,1))), np.array(sigma_e_NEV[:,:,:-1])), axis=-1)
                    )
            else:
                return

            return ErrorModel.Error(code, propagation, e_DIA, cov_DIA, e_NEV, e_NEV_star, sigma_e_NEV, cov_NEV)


    def drk_dDepth(self, survey):
        '''
        survey1 is previous survey station (with inc and azi in radians)
        survey2 is current survey station (with inc and azi in radians)
        '''
        # TODO: This is essentially minimum curvature... use function from utils instead (it's already in self.mc)
        md1, inc1, azi1 = np.array(survey[:-1]).T
        md2, inc2, azi2 = np.array(survey[1:]).T
        # delta_md = 1

        # dogleg = np.arccos(
        #     np.cos(inc2 - inc1)
        #     - (np.sin(inc1) * np.sin(inc2)) * (1 - np.cos(azi2 - azi1))
        # )
        
        # # manage discontinuity
        # with np.errstate(divide='ignore', invalid='ignore'):
        #     rf = np.ones_like(dogleg)
        #     rf = np.where(dogleg == 0, rf, 2 / dogleg * np.tan(dogleg / 2))
        
        # N = np.array(
        #     delta_md / 2 * (
        #     np.sin(inc1) * np.cos(azi1)
        #     + np.sin(inc2) * np.cos(azi2)
        #     ) * rf
        # )
                    
        # E = np.array(
        #     delta_md / 2 * (
        #     np.sin(inc1) * np.sin(azi1)
        #     + np.sin(inc2) * np.sin(azi2)
        #     ) * rf
        # )

        # V = np.array(
        #     delta_md / 2 * (
        #     np.cos(inc1) + np.cos(inc2)) * rf
        # )

        N = np.array(
            0.5 * (
            np.sin(inc1) * np.cos(azi1)
            + np.sin(inc2) * np.cos(azi2)
            )
        )

        E = np.array(
            0.5 * (
            np.sin(inc1) * np.sin(azi1)
            + np.sin(inc2) * np.sin(azi2)
            )
        )
        
        V = np.array(
            0.5 * (
            np.cos(inc1) + np.cos(inc2))
        )

        return np.vstack((np.array(np.zeros((1,3))), np.stack((N, E, V), axis=-1)))

    def drk_dInc(self, survey):
        '''
        survey1 is previous survey station (with inc and azi in radians)
        survey2 is current survey station (with inc and azi in radians)
        '''
        md1, inc1, azi1 = np.array(survey[:-1]).T
        md2, inc2, azi2 = np.array(survey[1:]).T
        delta_md = md2 - md1
        
        N = np.array(0.5 * ((delta_md) * np.cos(inc2) * np.cos(azi2)))
        E = np.array(0.5 * ((delta_md) * np.cos(inc2) * np.sin(azi2)))
        V = np.array(0.5 * (-delta_md * np.sin(inc2)))
        
        return np.vstack((np.array(np.zeros((1,3))), np.stack((N, E, V), axis=-1)))

    def drk_dAz(self, survey):
        '''
        survey1 is previous survey station (with inc and azi in radians)
        survey2 is current survey station (with inc and azi in radians)
        '''
        md1, inc1, azi1 = np.array(survey[:-1]).T
        md2, inc2, azi2 = np.array(survey[1:]).T
        delta_md = md2 - md1
        
        N = np.array(-0.5 * ((delta_md) * np.sin(inc2) * np.sin(azi2)))
        E = np.array(0.5 * ((delta_md) * np.sin(inc2) * np.cos(azi2)))
        V = np.zeros_like(N)
        
        return np.vstack((np.array(np.zeros((1,3))), np.stack((N, E, V), axis=-1)))

    def drkplus1_dDepth(self, survey):
        '''
        survey2 is current survey station (with inc and azi in radians)
        survey3 is next survey station (with inc and azi in radians)
        '''
        return np.vstack((self.drk_dDepth(survey)[1:] * -1, np.array(np.zeros((1,3)))))

    def drkplus1_dInc(self, survey):
        '''
        survey2 is current survey station (with inc and azi in radians)
        survey3 is next survey station (with inc and azi in radians)
        '''
        
        md2, inc2, azi2 = np.array(survey[:-1]).T
        md3, inc3, azi3 = np.array(survey[1:]).T
        delta_md = md3 - md2
        
        N = np.array(0.5 * ((delta_md) * np.cos(inc2) * np.cos(azi2)))
        E = np.array(0.5 * ((delta_md) * np.cos(inc2) * np.sin(azi2)))
        V = np.array(0.5 * (-(delta_md) * np.sin(inc2)))
        
        return np.vstack((np.stack((N, E, V), axis=-1), np.array(np.zeros((1,3)))))

    def drkplus1_dAz(self, survey):
        '''
        survey2 is current survey station (with inc and azi in radians)
        survey3 is next survey station (with inc and azi in radians)
        '''
        md2, inc2, azi2 = np.array(survey[:-1]).T
        md3, inc3, azi3 = np.array(survey[1:]).T
        delta_md = md3 - md2
        
        N = np.array(-0.5 * ((delta_md) * np.sin(inc2) * np.sin(azi2)))
        E = np.array(0.5 * ((delta_md) * np.sin(inc2) * np.cos(azi2)))
        V = np.zeros_like(N)
        
        return np.vstack((np.stack((N, E, V), axis=-1), np.array(np.zeros((1,3)))))

    def _drdp(self, survey):
        
        return np.hstack((
            self.drk_dDepth(survey),
            self.drk_dInc(survey),
            self.drk_dAz(survey),
            self.drkplus1_dDepth(survey),
            self.drkplus1_dInc(survey),
            self.drkplus1_dAz(survey)
        ))

    def _drdp_sing(self, survey):
        '''
        survey1 is previous survey station (with inc and azi in radians)
        survey2 is current survey station
        survey3 is next survey station (with inc and azi in radians)
        '''
        md1, inc1, azi1 = np.array(survey[:-2]).T
        md2, inc2, azi2 = np.array(survey[1:-1]).T
        md3, inc3, azi3 = np.array(survey[2:]).T
        double_delta_md = md3 - md1
        delta_md = md2 - md1

        return dict(
            double_delta_md = double_delta_md,
            delta_md = delta_md,
            azi2 = azi2
        )





import json
import numpy as np
import pandas as pd
from welleng.error import ErrorModel
from welleng.utils import get_sigmas

"""
Test that the ISCWSA MWD Rev4 error model is working within a defined
tolerance (the default has been set to 0.001%), testing against the
MWD Rev 4 error model example provided by ISCWSA.
"""

# Set test tolerance as percentage
TOLERANCE = 0.001

# Read validation data from file:
wd = json.load(open("test/test_data/error_mwdrev4_iscwsa_well_data.json"))
vd = json.load(open("test/test_data/error_mwdrev4_iscwsa_validation.json"))

df = pd.DataFrame(vd)


def get_md_index(error_data, md):
    i = np.where(error_data.survey_deg == md)[0][0]
    return i


# generate error model
err = ErrorModel(
    survey=wd['survey'],
    well_ref_params=wd['well_ref_params'],
    AzimuthRef=wd['well_ref_params']['AzimuthRef']
)


# initiate lists
def test_error_model_mwdrev4_iscwsa(df=df, err=err):
    nn_c, ee_c, vv_c, ne_c, nv_c, ev_c = [], [], [], [], [], []
    data = [
        nn_c, ee_c, vv_c, ne_c, nv_c, ev_c
    ]

    # generate error data
    for index, row in df.iterrows():
        i = get_md_index(err, row['md'])
        s = row['source'].replace('-', '_')
        if s == "Totals":
            source_cov = err.errors.cov_NEVs.T[i]
        else:
            source_cov = err.errors.errors[s].cov_NEV.T[i]
        v = get_sigmas(source_cov, long=True)
        for j, d in enumerate(v):
            data[j].append(d[0])

    # convert to dictionary
    ed = {}
    headers = [
        'nn_c', 'ee_c', 'vv_c', 'ne_c', 'nv_c', 'ev_c'
    ]
    for i, h in enumerate(headers):
        ed[h] = data[i]

    df_c = pd.DataFrame(ed)

    df_r = df.join(df_c)

    headers = [
        'nn_d', 'ee_d', 'vv_d', 'ne_d', 'nv_d', 'ev_d'
    ]
    df_d = pd.DataFrame(
        np.around(
            np.array(df_c) - np.array(df.iloc[:, 2:]),
            decimals=4
        ),
        columns=headers
    )

    df_r = df_r.join(df_d)

    with np.errstate(divide='ignore', invalid='ignore'):
        error = np.nan_to_num(np.absolute(
            np.array(df_d) / np.array(df.iloc[:, 2:])
            ) * 100
        )

    assert np.all(error < TOLERANCE)

    # if you wanted to view the results, this would save then to an Excel
    # file.
    # df_r.to_excel("test/test_data/error_mwdrev4_iscwsa_validation_results.xlsx")


# make above test runnanble separately
if __name__ == '__main__':
    test_error_model_mwdrev4_iscwsa(df=df, err=err)
